###############################################################
###    IPG - IDSIA
###    Authors: L. Azzimonti, G. Corani
###    Reference: "Hierarchical Multinomial-Dirichlet model for the estimation of conditional probability tables",
###    Proceedings of 2017 IEEE International Conference on Data Mining (ICDM 2017)
###    Date: August 22, 2017
###
###    This code is for demostration purpose only
###    for licensing contact laura@idsia.ch
###    COPYRIGHT (C) 2017 IDSIA
###############################################################

library(bnlearn)
source("src/VAHierTest.R")
source("src/convert2array.R")
source("src/compute_loglike.R")
source("src/compute_mse.R")
source("src/predict.R")
source("src/compute_theta_MD_mean.R")


args=c(11,3,"nursery")
#args[1] = repetition of the experiment
#args[2] = number of observation in training set. Possible values: 1-7
#args[3] = name of the dataset. Possible values: "adult", "letter", "nursery", "pendigits", "spambase".

repetitions=1 
set_repetitions=as.numeric(args[1])
id_n_vec=as.numeric(args[2])
n_vec=10*(2^(1:11))[id_n_vec]

name_dataset=args[3]
source(paste("dataset/",name_dataset,".R",sep=""))

pdag = tree.bayes(dati, class, features_evidence)
plot(pdag)
dag=pdag

fit_true = bn.fit(dag, dati, method = "bayes")
fit_true
fit_true_MLE = bn.fit(dag, dati)
data=NULL


MSE_mat=NULL
MSE_hier_mat=NULL
MSE_MLE_mat=NULL
MSE_bayes1_mat=NULL
MSE_bayes10_mat=NULL
ll_hier_mat=NULL
ll_bayes1_mat=NULL
ll_bayes10_mat=NULL
ll_MLE_mat=NULL
ll_mat=NULL
llR_mat=NULL
acc_mat=NULL
log_loss_mat=NULL
dati_pred_hier=NULL
dati_pred_bayes1=NULL
dati_pred_bayes10=NULL
dati_pred_MLE=NULL
cont_pred_hier=NULL
cont_pred_bayes1=NULL
cont_pred_bayes10=NULL
cont_pred_MLE=NULL
ROC_mat=NULL

for(sim in 1:repetitions){
  for(n in n_vec){
    id=sample(1:N,n)
    id_test=sample(1:(N-n),max(0,min(1000,(N-n))))
    dati_test=dati[-id,][id_test,]
    
    fit_MLE = bn.fit(dag, dati[id,])
    fit_bayes1 = bn.fit(dag, dati[id,], method = "bayes",iss=1)
    fit_bayes10 = bn.fit(dag, dati[id,], method = "bayes",iss=10)
    
    stanFitShrinkage=NULL
    prob_true=NULL
    prob_hier=NULL
    prob_hier_array=NULL
    meaning_zStates_list=NULL
    MSE_hier=NULL
    MSE_MLE=NULL
    MSE_bayes1=NULL
    MSE_bayes10=NULL
    done=list()
    prob_bayes1_array=list()
    prob_bayes10_array=list()
    prob_MLE_array=list()
    
    for( i in 1:p){
      if(length(dag$nodes[[i]]$parents)>0){
        data=NULL
        data$xValues=as.numeric(dati[,variables[i]])[id]
        levelsx=levels(dati[,variables[i]])
        data$xValues=as.numeric(dati[id,variables[i]])
        
        newstates_all=dati[,dag$nodes[[i]]$parents[length(dag$nodes[[i]]$parents)]]
        if(length(dag$nodes[[i]]$parents)>1){
          for(j in (length(dag$nodes[[i]]$parents)-1):1){
            newstates_all=paste0(newstates_all, "/", dati[,dag$nodes[[i]]$parents[j]])
          }
        }
        meaning_zStates=levels(as.factor(newstates_all))
        
        newstates=newstates_all
        data$zValues=as.numeric(as.factor(newstates))[id]
        
        I=length(levelsx)
        J=length(levels(as.factor(newstates)))
        
        ris_hier=VAHierTest(data$xValues, data$zValues,I, J,stanFitShrinkage)
        sample_theta=ris_hier$theta
        prob_hier[[variables[i]]]=matrix(colMeans(sample_theta),I,J)
        meaning_zStates_list[[variables[i]]]=meaning_zStates
        prob_hier_array[[variables[i]]]=convert2array(prob_hier[[variables[i]]],meaning_zStates_list[[variables[i]]],dati,variables,dag,I,J)
        
        done[[variables[i]]]=1
      }else{
        data=NULL
        data$xValues=as.numeric(dati[,variables[i]])[id]
        levelsx=levels(dati[,variables[i]])
        I=length(levelsx)
        data$xCounts=as.numeric(table(factor(data$xValues,levels=1:I)))
        theta=compute_theta_MD_mean(data$xCounts,I)
        prob_hier_array[[variables[i]]]=theta
        done[[variables[i]]]=1
      }
      mse_mean_hier=compute_mse_mean(prob_hier_array[[variables[i]]], fit_true[[variables[i]]]$prob )
      mse_mean_MLE=compute_mse_mean(fit_MLE[[variables[i]]]$prob, fit_true[[variables[i]]]$prob )
      mse_mean_bayes1= compute_mse_mean(fit_bayes1[[variables[i]]]$prob, fit_true[[variables[i]]]$prob )
      mse_mean_bayes10= compute_mse_mean(fit_bayes10[[variables[i]]]$prob, fit_true[[variables[i]]]$prob )
      
      MSE_hier=c(MSE_hier,mse_mean_hier)
      MSE_MLE=c(MSE_MLE,mse_mean_MLE)
      MSE_bayes1=c(MSE_bayes1,mse_mean_bayes1)
      MSE_bayes10=c(MSE_bayes10,mse_mean_bayes10)
      
      prob_bayes1_array[[variables[i]]]=fit_bayes1[[variables[i]]]$prob
      prob_bayes10_array[[variables[i]]]=fit_bayes10[[variables[i]]]$prob
      prob_MLE_array[[variables[i]]]=fit_MLE[[variables[i]]]$prob
    }
    
    names(MSE_hier)=variables
    names(MSE_MLE)=variables
    names(MSE_bayes1)=variables
    names(MSE_bayes10)=variables
    MSE_hier_mat=rbind(MSE_hier_mat,c(MSE_hier,N=n))
    MSE_MLE_mat=rbind(MSE_MLE_mat,c(MSE_MLE,N=n))
    MSE_bayes1_mat=rbind(MSE_bayes1_mat,c(MSE_bayes1,N=n))
    MSE_bayes10_mat=rbind(MSE_bayes10_mat,c(MSE_bayes10,N=n))
    MSE_mat=rbind(MSE_mat,c(MLE=mean(MSE_MLE), BAYES_1=mean(MSE_bayes1),BAYES_10=mean(MSE_bayes10),HIER=mean(MSE_hier) ,N=n))
    
    ll_hier=compute_loglike(dati_test,prob_hier_array)
    ll_bayes1=compute_loglike(dati_test,prob_bayes1_array)
    ll_bayes10=compute_loglike(dati_test,prob_bayes10_array)
    ll_MLE=compute_loglike(dati_test,prob_MLE_array)
    
    ll_hier_mat=ll_hier$logPvec
    ll_bayes1_mat=ll_bayes1$logPvec
    ll_bayes10_mat=ll_bayes10$logPvec
    ll_MLE_mat=ll_MLE$logPvec
    ll_all=cbind(ll_MLE$logPvec, ll_bayes1$logPvec, ll_bayes10$logPvec,ll_hier$logPvec)
    colnames(ll_all)=c("MLE","BAYES1","BAYES10","HIER")
    
    sorted_ll=ll_all[sort(ll_all[,"HIER"], index.return=TRUE)$ix,]
    
    ll_mat=rbind(ll_mat,c(MLE=ll_MLE$logP, BAYES_1=ll_bayes1$logP,BAYES_10=ll_bayes10$logP,HIER=ll_hier$logP ,N=n))
    llR_mat=rbind(llR_mat,c(HIER_vs_MLE=ll_hier$logP-ll_MLE$logP, HIER_vs_BAYES_1=ll_hier$logP-ll_bayes1$logP,HIER_vs_BAYES_10=ll_hier$logP-ll_bayes10$logP,N=n))
    
    pred_hier=predict2(dati_test,prob_hier_array, class)
    pred_bayes1=predict2(dati_test,prob_bayes1_array, class)
    pred_bayes10=predict2(dati_test,prob_bayes10_array, class)
    
    dati_pred_hier=rbind(dati_pred_hier,pred_hier$dati_pred)
    dati_pred_bayes1=rbind(dati_pred_bayes1,pred_bayes1$dati_pred)
    dati_pred_bayes10=rbind(dati_pred_bayes10,pred_bayes10$dati_pred)
    
    cont_pred_hier=rbind(cont_pred_hier,as.numeric(pred_hier$cont)) #bycol (col1, col2, col3)
    cont_pred_bayes1=rbind(cont_pred_bayes1,as.numeric(pred_bayes1$cont))
    cont_pred_bayes10=rbind(cont_pred_bayes10,as.numeric(pred_bayes10$cont))
    
    acc_mat=rbind(acc_mat,c(NA,pred_bayes1$acc, pred_bayes10$acc,pred_hier$acc,N=n))#,nfold=n_fold[sim]))
    log_loss_mat=rbind(log_loss_mat,c(NA,pred_bayes1$log_loss, pred_bayes10$log_loss,pred_hier$log_loss,N=n))#, nfold=n_fold[sim]))
    ROC_mat=rbind(ROC_mat,c(NA,pred_bayes1$ROC, pred_bayes10$ROC,pred_hier$ROC,N=n))#,nfold=n_fold[sim]))
  }
}

cont_all= cbind(rep(NA, dim(cont_pred_bayes1)[2]), colSums(cont_pred_bayes1), colSums(cont_pred_bayes10),colSums(cont_pred_hier))
dati_pred_all=list()
dati_pred_all[["MLE"]]=dati_pred_MLE
dati_pred_all[["BAYES_1"]]=dati_pred_bayes1
dati_pred_all[["BAYES_10"]]=dati_pred_bayes10
dati_pred_all[["HIER"]]=dati_pred_hier

colnames(ll_mat)=c("MLE", "BAYES_1", "BAYES_10","HIER", "N")
colnames(llR_mat)=c("HIER_vs_MLE", "HIER_vs_BAYES_1", "HIER_vs_BAYES_10", "N")
colnames(acc_mat)=c("MLE", "BAYES_1", "BAYES_10","HIER", "N")
colnames(log_loss_mat)=c("MLE", "BAYES_1", "BAYES_10","HIER", "N")
colnames(cont_all)=c("MLE", "BAYES_1", "BAYES_10","HIER")
colnames(ROC_mat)=c("MLE", "BAYES_1", "BAYES_10","HIER", "N")

save(file=paste("ris_",name_dataset,"_tan_n_",id_n_vec, "_rep_",set_repetitions,".Rdata",sep=""),MSE_mat, MSE_hier_mat,MSE_MLE_mat, MSE_bayes1_mat, MSE_bayes10_mat, ll_mat, llR_mat, acc_mat,log_loss_mat,cont_all,ROC_mat)
