###############################################################
###    IPG - IDSIA
###    Authors: L. Azzimonti, G. Corani
###    Reference: "Hierarchical Multinomial-Dirichlet model for the estimation of conditional probability tables",
###    Proceedings of 2017 IEEE International Conference on Data Mining (ICDM 2017)
###    Date: August 22, 2017
###
###    This code is for demostration purpose only
###    for licensing contact laura@idsia.ch
###    COPYRIGHT (C) 2017 IDSIA
###############################################################

predict_singleD=function(D,prob_array,class){
    if(length(dag$nodes[[class]]$parents)==0){
      Pclass=prob_array[[class]]
    }else{
      stringa="Pclass=prob_array[[class]]["
      for(j in 1:length(dag$nodes[[class]]$parents)){
        k=as.numeric(D[dag$nodes[[class]]$parents[j]])
        stringa=paste(stringa, k, sep=",")
      }
      stringa=paste(stringa, "]", sep="")
      eval(parse(text=stringa))
    }
      
  return(Pclass)
}


predict=function(dati_test,prob_array,class){
  P=NULL
  Ptrue=NULL
  for(Drow in 1:(dim(dati_test)[1])){
    D=dati_test[Drow,]
    Ptemp=predict_singleD(D,prob_array,class)
    P=rbind(P,Ptemp)
    Ptrue=c(Ptrue,Ptemp[as.numeric(dati_test[Drow,class])])
  }
  lev=levels(dati_test[,class])
  p_most_prob=apply(P,1,max)
  if(sum(is.nan(p_most_prob))==0){
    most_prob=apply(P,1,which.max)
  }else{
    most_prob=rep(NaN,dim(P)[1])
    most_prob[!is.nan(p_most_prob)]=apply(P[!is.nan(p_most_prob),],1,which.max)
  }
  dati_pred=data.frame(true=dati_test[,class], pred=lev[most_prob],p_true=Ptrue, p_pred=p_most_prob, P ,row.names = NULL)
  dati_pred[,2]=factor(dati_pred[,2],levels=levels(dati_pred[,1]))
  acc=mean(dati_pred[,1]==dati_pred[,2])
  cont=table(dati_pred[,1],dati_pred[,2])
  log_loss=-sum(log(dati_pred[,3]))
  ROC=wilcox.test(dati_pred$p_true[dati_pred$true==lev[1]],dati_pred$p_true[dati_pred$true==lev[2]])$statistic
  ROC=ROC/prod(rowSums(cont))
  return(list(dati_pred=dati_pred,acc=acc, cont=cont, log_loss=log_loss, ROC=ROC))
}




predict2_singleD=function(Dmat,prob_array,class){
  log_P_vec=NULL
  for(row in 1:(dim(Dmat)[1])){
    D=Dmat[row,]
    log_P=0
    for(i in 1:p){
      if(length(dag$nodes[[i]]$parents)==0){
        logP_temp=log(prob_array[[variables[i]]][as.numeric(D[variables[i]])])
      }else{
        stringa="logP_temp=log(prob_array[[variables[i]]]["
        k=as.numeric(D[variables[i]])
        stringa=paste(stringa, k, sep="")
        for(j in 1:length(dag$nodes[[i]]$parents)){
          k=as.numeric(D[dag$nodes[[i]]$parents[j]])
          stringa=paste(stringa, k, sep=",")
        }
        stringa=paste(stringa, "])", sep="")
        eval(parse(text=stringa))
      }
      log_P=log_P+logP_temp
    }
    log_P_vec=c(log_P_vec,log_P)
  }
  P=exp(log_P_vec)
  P=P/sum(P)
  return(P)
}


predict2=function(dati_test,prob_array,class){
  P=NULL
  Ptrue=NULL
  for(Drow in 1:(dim(dati_test)[1])){
    class_true=dati_test[Drow,class]
    lev=levels(dati_test[,class])
    Dmat=NULL
    for(v in lev){
      dati_test[Drow,class]=v
      Dmat=rbind(Dmat,dati_test[Drow,])
    }
    dati_test[Drow,class]=class_true
    Ptemp=predict2_singleD(Dmat,prob_array,class)
    P=rbind(P,Ptemp)
    Ptrue=c(Ptrue,Ptemp[as.numeric(dati_test[Drow,class])])
  }
  lev=levels(dati_test[,class])
  p_most_prob=apply(P,1,max)
  if(sum(is.nan(p_most_prob))==0){
    most_prob=apply(P,1,which.max)
  }else{
    most_prob=rep(NaN,dim(P)[1])
    most_prob[!is.nan(p_most_prob)]=apply(P[!is.nan(p_most_prob),],1,which.max)
  }
  
  dati_pred=data.frame(true=dati_test[,class], pred=lev[most_prob],p_true=Ptrue, p_pred=p_most_prob, P ,row.names = NULL)
  dati_pred[,2]=factor(dati_pred[,2],levels=levels(dati_pred[,1]),ordered=is.ordered(dati_pred[,1]))
  acc=mean(dati_pred[,1]==dati_pred[,2])
  
  names(dati_pred)[5:length(names(dati_pred))]=lev
  cont=table(dati_pred[,1],dati_pred[,2])
  
  log_loss=-sum(log(dati_pred[,3]))

  ROC_vec=NULL
  for(l in lev){
    if(rowSums(cont)[l]>0){
      ROC_vec=c(ROC_vec,wilcox.test(dati_pred[dati_pred$true==l,l],dati_pred[dati_pred$true!=l,l])$statistic/(rowSums(cont)[l]*(sum(cont)-rowSums(cont)[l])))
    }else{
      ROC_vec=c(ROC_vec,0)
    }
  }
  ROC=sum(ROC_vec*rowSums(cont)/sum(cont))
  
  return(list(dati_pred=dati_pred,acc=acc, cont=cont, log_loss=log_loss,ROC=as.numeric(ROC)))
}

